@@ -5,7 +5,7 @@ from .lr_scheduler_wrapper import LRSchedulerWrapper
 from .metrics_logger import MetricsLoggingMode, TrainingMetricsLogger, TestMetricsLogger
 from .tensorboard import Tensorboard, AutoTensorboard
 from .distributed import DistributedSamplerManager
-from .cuda_cache_management import EmptyCudaCache
+from .cuda_memory import EmptyCudaCache, LogCudaMemory
 
 
 
