@@ -10,8 +10,14 @@ import basics.base_utils as _
 
 class LogProgress(Callback):
 
-    def __init__(self, log_period=200, set_names=None, batch_level=True, logs_base_path="current", name="LogProgress"):
-        super(LogProgress, self).__init__(name=name)
+    def __init__(self,
+                 log_period=200,
+                 set_names=None,
+                 batch_level=True,
+                 logs_base_path="current",
+                 name="LogProgress",
+                 **kwargs):
+        super(LogProgress, self).__init__(name=name, **kwargs)
 
         self.log_period = log_period
         self.set_names = set_names or ["training"]
@@ -45,34 +51,37 @@ class LogProgress(Callback):
             eta = self._calc_eta(logs)
             average_duration = self._get_average_batch_duration(logs)
 
-            sys.stdout.write('\nEpoch {:d}/{:d} - ETA: {:s}\tBatch {:d}/{:d} '
-                             'Average batch training time {:s}\n'.format(current["epoch"],
-                                                                         logs["final_epoch"],
-                                                                         eta,
-                                                                         current["batch_step"],
-                                                                         logs["final_batch_step"],
-                                                                         average_duration))
+            self._write('\nEpoch {:d}/{:d} - ETA: {:s}\tBatch {:d}/{:d} '
+                        'Average batch training time {:s}\n'.format(current["epoch"],
+                                                                    logs["final_epoch"],
+                                                                    eta,
+                                                                    current["batch_step"],
+                                                                    logs["final_batch_step"],
+                                                                    average_duration))
 
             for metric_level in ['batch', 'window_average', 'dataset', 'epoch']:
                 self._write_metric_logs(metric_level, logs)
-                sys.stdout.write(f'\n')
+                self._write(f'\n')
 
-            sys.stdout.write(f'\n')
+            self._write(f'\n')
 
         return success
 
     def on_epoch_completed(self, logs):
         current = self._get_logs_base(logs)
         duration = self._get_epoch_duration(logs)
-        sys.stdout.write('\nEpoch {:d}/{:d}\tREADY - Duration {:s}\n'.format(current["epoch"],
+        self._write('\n')
+        self._write('###############################################################################')
+        self._write('\n')
+        self._write('Epoch {:d}/{:d}\tREADY - Duration {:s}\n'.format(current["epoch"],
                                                                            logs["final_epoch"],
                                                                            duration))
         success = True
         for metric_level in ['window_average', 'dataset', 'epoch']:
             self._write_metric_logs(metric_level, logs)
-            sys.stdout.write(f'\n')
+            self._write(f'\n')
 
-        sys.stdout.write(f'\n')
+        self._write(f'\n')
 
         return success
 
@@ -136,8 +145,8 @@ class LogProgress(Callback):
             metrics_log += f'{set_name:<15}: {set_metrics_log}.\n'
 
         if len(metrics_log) > 0:
-            sys.stdout.write(f'{self.metric_level_names[metric_level]}:\n')
-            sys.stdout.write(metrics_log)
+            self._write(f'{self.metric_level_names[metric_level]}:\n')
+            self._write(metrics_log)
 
     def _create_set_metrics_log_for(self, set_name, metric_level, logs):
         current = self._get_logs_base(logs)
@@ -162,20 +171,28 @@ class LogProgress(Callback):
         if base_metric is not None:
             log += f"{base_metric:<15}: "
 
+        metric_value_logs = []
         for c, (metric, value) in enumerate(metrics.items()):
             if metric in skip_metric_names:
                 continue
 
-            if type(value) is dict:
-                log += "\n" + self._create_log_for(value, metric, log_depth+1)
-            elif isinstance(value, (float, int)):
-                log_format = self._get_log_format(value)
-                log += log_format.format(metric, value)
-            else:
-                log += "[UNKNOWN]"
+            if type(value) is tuple:
+                # use the first value as metric value, the other values are auxiliary results meant for other purposes
+                value = value[0]
 
-            if c < num_metrics - 1:
-                log += ', '
+            if type(value) is dict:
+                nested_logs = self._create_log_for(value, metric, log_depth+1)
+                if nested_logs is not None:
+                    metric_value_logs += ["\n" + nested_logs]
+            else:
+                try:
+                    log_format = self._get_log_format(value)
+                    metric_value_logs += [log_format.format(metric, value)]
+                except Exception as e:
+                    metric_value_logs += ["[UNKNOWN]"]
+
+        if len(metric_value_logs) > 0:
+            log += ', '.join(metric_value_logs)
 
         return log
 
@@ -187,11 +204,15 @@ class LogProgress(Callback):
 
         return log_format
 
+    def _write(self, text):
+        sys.stdout.write(text)
+        sys.stdout.flush()
+
 
 class BatchSizeLogger(Callback):
 
-    def __init__(self, batch_dimension=1, name="BatchSizeLogger"):
-        super().__init__(name=name)
+    def __init__(self, batch_dimension=1, name="BatchSizeLogger", **kwargs):
+        super().__init__(name=name, **kwargs)
 
         self._batch_dimension = batch_dimension
 
@@ -206,6 +227,7 @@ class BatchSizeLogger(Callback):
 
         current = self._get_logs_base(logs)
 
+        # TODO : doesn't work for Tensorflow
         current['training_params']['batch']['batch_size'] = len(training_batch) if is_chunkable(training_batch) else \
             training_batch[0].size(self._batch_dimension)
 
