@@ -1,9 +1,12 @@
 import os
+import sys
 import time
 import json
 
 import abc
 
+import pickle
+
 from shutil import copyfile
 
 from mlpug.utils import get_value_at
@@ -12,7 +15,7 @@ from mlpug.trainers.callbacks.callback import Callback
 import basics.base_utils as _
 
 
-class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
+class CheckpointManager(Callback, metaclass=abc.ABCMeta):
 
     def __init__(self,
                  model_hyper_parameters=None,
@@ -26,11 +29,11 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                  archive_last_model_checkpoint_every=2000,  # batches or epochs
                  force_monitoring_on_epoch=True,
                  base_checkpoint_filename=time.strftime("%d-%m-%Y_%H-%M-%S"),
-                 model_checkpoint_filename_ext="model",
-                 training_checkpoint_filename_ext="state",
+                 model_checkpoint_filename_ext="m-ckp",
+                 training_checkpoint_filename_ext="t-ckp",
                  backup_before_override=True,
-                 disable_saving_checkpoints=False,
-                 name="CheckpointManager"):
+                 name="CheckpointManager",
+                 **kwargs):
         """
 
         :param model_hyper_parameters: Dict with hyper parameters of the model, this will be saved as part of
@@ -49,8 +52,13 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
 
         :param create_checkpoint_every: period before saving next training/model checkpoint
                                         (will be overridden after each period)
+                                        A 0 value disables this feature.
+
         :param archive_last_model_checkpoint_every: period before the last available model checkpoint is archived.
                                                     Period must be multiple of create_checkpoint_every period
+
+                                                    A 0 value disables this feature.
+
         :param force_monitoring_on_epoch: When True, the given metric will also be monitored on every epoch
                                           in the case that monitoring level is batch level
         :param base_checkpoint_filename:
@@ -60,10 +68,8 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                                        overridden by a new version. If backing up fails, no new version will be written
                                        to disk. This gives the user a chance to fix a problem, e.g. disk full, without
                                        interruption of the training process.
-        :param disable_saving_checkpoints: When True no checkpoints at all will be saved, e.g. for testing training
-
         """
-        super().__init__(name=name)
+        super().__init__(name=name, **kwargs)
 
         self._model_hyper_parameters = model_hyper_parameters
 
@@ -82,15 +88,13 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
         self._create_checkpoint_every = create_checkpoint_every
         self._archive_last_model_checkpoint_every = archive_last_model_checkpoint_every
 
-        self._disable_saving_checkpoints = disable_saving_checkpoints
-
         self._force_monitoring_on_epoch = force_monitoring_on_epoch
 
+        self._backup_before_override = backup_before_override
+
         self._best_model_quality = float('Inf') if self._metric_opt_mode == 'min' else -float('Inf')
         self._best_model_iter = None
 
-        self._backup_before_override = backup_before_override
-
         self._check_settings()
 
         self._describe_setup()
@@ -130,7 +134,6 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
         return self._monitor('global_iter', logs)
 
     def on_epoch_completed(self, logs):
-        force = False
         iter_name = 'epoch'
 
         force_monitoring = False
@@ -144,13 +147,40 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
 
         return self._monitor(iter_name, logs, force_monitoring=force_monitoring)
 
-    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
-        if abs(self._best_model_quality) == float('Inf') or self._best_model_iter is None:
-            return True
+    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
+        status = 'ended'
+        if stopped_early:
+            status = 'stopped early'
 
-        self._log.info(f"Best model quality reached {self._metric_to_monitor}={self._best_model_quality} "
-                       f"at global iteration {self._best_model_iter}")
-        return True
+        if stopped_on_error:
+            status = 'stopped on error'
+
+        if interrupted:
+            status = 'interrupted'
+
+        sys.stdout.write("\n")
+        sys.stdout.flush()
+        self._log.info(f"Training {status}")
+        success = True
+        if self._create_checkpoint_every > 0:
+            self._log.info(f"... storing latest model ...")
+            latest_model_fname = self._create_model_checkpoint()
+            success = (latest_model_fname is not None)
+
+        if stopped_early or interrupted:
+            self._log.info(f"... storing training checkpoint ...")
+            training_checkpoint_fname = self._create_training_checkpoint()
+            success &= (training_checkpoint_fname is not None)
+
+        if abs(self._best_model_quality) != float('Inf') and self._best_model_iter is not None:
+            iter_name = "global iteration" if self._batch_level else "epoch"
+
+            self._log.info(f"Best model quality reached {self._metric_to_monitor}={self._best_model_quality} "
+                           f"at {iter_name} {self._best_model_iter}")
+        else:
+            self._log.warn(f"No good enough model found, no best model checkpoint saved.")
+
+        return success
 
     def training_checkpoint_file_name(self):
         b_fn = self._base_filename
@@ -187,7 +217,13 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                        f"{self._archive_last_model_checkpoint_every} {time_scale}")
 
     def _get_model_quality(self, current_logs):
-        return get_value_at(self._metric_to_monitor, current_logs)
+        model_quality = get_value_at(self._metric_to_monitor, current_logs)
+
+        if type(model_quality) is tuple:
+            # use the first value as metric value, the other values are auxiliary results meant for other purposes
+            model_quality = model_quality[0]
+
+        return model_quality
 
     # TODO : it would maybe be better to split the monitoring and the checkpointing in to two separated methods
     def _monitor(self, iter_name, logs, force_monitoring=False):
@@ -195,8 +231,7 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
 
         training_iter = current[iter_name]
 
-        if not force_monitoring and training_iter == 0:
-            return True
+        pretty_iter_name = self.pretty_iter_name(iter_name)
 
         best_model_fname = None
         success = True
@@ -214,32 +249,34 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                     model_quality_good_enough = model_quality >= self._metric_checkpoint_threshold
 
             if not model_quality_good_enough:
-                self._log.debug("Iter : %s : model quality not good enough : %s : %3e " % (training_iter,
-                                                                                           self._metric_to_monitor,
-                                                                                           model_quality))
+                self._log.debug("%s : %s : model quality not good enough : %s : %3e " % (pretty_iter_name,
+                                                                                         training_iter,
+                                                                                         self._metric_to_monitor,
+                                                                                         model_quality))
 
             if model_quality_valid and model_quality_good_enough:
                 model_improved = ((self._metric_opt_mode == 'min') and (model_quality < self._best_model_quality)) or \
                                  ((self._metric_opt_mode == 'max') and (model_quality > self._best_model_quality))
 
                 if model_improved:
-                    self._log.debug("Iter : %s : Model improved : %s : %3e " % (training_iter,
-                                                                                self._metric_to_monitor,
-                                                                                model_quality))
+                    self._log.debug("%s : %s : Model improved : %s : %3e " % (pretty_iter_name,
+                                                                              training_iter,
+                                                                              self._metric_to_monitor,
+                                                                              model_quality))
 
                     if (self._best_model_iter is not None) and \
                             (iter_name == 'epoch') and \
                             (training_iter < self._best_model_iter):
-                        self._log.warn(f"Inconsistency: according to the current training iter ({training_iter}), "
-                                       f"current best model training iter ({self._best_model_iter}) is in the future. "
-                                       f"Was the right training checkpoint loaded?")
+                        self._log.warn(f"Inconsistency: according to the current training {pretty_iter_name} "
+                                       f"({training_iter}), current best model training iter ({self._best_model_iter}) "
+                                       f"is in the future. Was the right training checkpoint loaded?")
 
-                    best_model_fname = self._disable_saving_checkpoints or self._save_current_model_as_best()
+                    best_model_fname = self._save_current_model_as_best()
                     if best_model_fname:
                         self._best_model_quality = model_quality
                         self._best_model_iter = training_iter
 
-                        data_saved = not self._disable_saving_checkpoints
+                        data_saved = True
                     else:
                         self._log.error("Unable to save improved model checkpoint")
                         success = False
@@ -250,9 +287,7 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                                 f"skipping ...")
                 success = False
 
-        if not self._disable_saving_checkpoints and \
-                (self._create_checkpoint_every > 0) and \
-                (training_iter % self._create_checkpoint_every == 0):
+        if (self._create_checkpoint_every > 0) and (training_iter % self._create_checkpoint_every == 0):
             # Just copy best model if available
             latest_model_fname = self._create_model_checkpoint(file_to_copy=best_model_fname)
 
@@ -273,7 +308,8 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                     self._log.error("Unable to create checkpoint for latest model, unable to archive latest model")
                     success = False
 
-            training_checkpoint_success = self._create_training_checkpoint()
+            training_checkpoint_fname = self._create_training_checkpoint()
+            training_checkpoint_success = (training_checkpoint_fname is not None)
             data_saved |= training_checkpoint_success
 
             success &= training_checkpoint_success
@@ -284,7 +320,6 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
         return success
 
     def _save_current_model_as_best(self):
-
         model_fn = self.best_model_file_name()
 
         if self._backup_before_override:
@@ -295,13 +330,7 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                                            f"will not override override model checkpoint with new one", e)
                 return None
 
-        try:
-            self._save_model_checkpoint(model_fn)
-        except Exception as e:
-            _.log_exception(self._log, f"A problem occurred saving the current model as best model", e)
-            return None
-
-        return model_fn
+        return self._create_model_checkpoint(model_fn=model_fn)
 
     def _update_logs(self, model_improved, logs, current):
         if model_improved:
@@ -309,15 +338,26 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
 
         current["is_best"] = model_improved
 
-    def _create_model_checkpoint(self, file_to_copy=None):
-        model_fn = self.latest_model_file_name()
+    def _create_model_checkpoint(self, model_fn=None, file_to_copy=None):
+        if model_fn is None:
+            model_fn = self.latest_model_file_name()
+
         if file_to_copy:
             if not self._copy(file_to_copy, model_fn):
                 self._log.error(f"Unable to create model checkpoint based on file : {file_to_copy}")
                 return None
         else:
             try:
-                self._save_model_checkpoint(model_fn)
+                state, success = self._gather_model_checkpoint_data()
+                if state is not None:
+                    if not success:
+                        self._log.warn("Gathering the model checkpoint data was not completely successful, "
+                                       "will save available checkpoint data anyway ...")
+
+                    self._log.debug(f"Saving model checkpoint : {model_fn}")
+                    self._save_model_checkpoint(model_fn, state)
+                else:
+                    return None
             except Exception as e:
                 _.log_exception(self._log, f"A problem occurred saving the latest model as checkpoint", e)
                 return None
@@ -332,27 +372,77 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                 if not self._backup_checkpoint(checkpoint_fname):
                     self._log.error("Unable to backup last training checkpoint, "
                                     "will not override override training checkpoint with new one")
-                    return False
+                    return None
             except Exception as e:
                 _.log_exception(self._log, f"A problem occurred backing up the last training checkpoint, "
                                            f"will not override override training checkpoint with new one", e)
-                return False
+                return None
 
-        success = False
         try:
-            success = self._save_training_checkpoint(checkpoint_fname)
+            state, success = self._gather_training_checkpoint_data()
+            if state is not None:
+                if not success:
+                    self._log.warn("Gathering the training checkpoint data was not completely successful, "
+                                   "will save available checkpoint data anyway ...")
+
+                self._log.debug(f"Saving training checkpoint : {checkpoint_fname}")
+                self._save_training_checkpoint(checkpoint_fname, state)
+            else:
+                return None
         except Exception as e:
             _.log_exception(self._log, f"Unable to save training checkpoint", e)
+            return None
 
-        return success
+        return checkpoint_fname
 
-    @abc.abstractmethod
-    def _save_training_checkpoint(self, filename):
-        self._log.error("This method is not implemented, implement it in your child class implementation")
+    def _gather_model_checkpoint_data(self):
+        """
 
-    @abc.abstractmethod
-    def _save_model_checkpoint(self, filename):
-        self._log.error("This method is not implemented, implement it in your child class implementation")
+        :return: state, success
+        """
+        state, success = self.trainer.get_model_components_state()
+
+        if state is not None:
+            if not success:
+                self._log.warn("Getting the model components state was not completely successful, "
+                               "continuing anyway ...")
+
+            if self._model_hyper_parameters is not None:
+                state['hyper_parameters'] = self._model_hyper_parameters
+
+            try:
+                manager_state, manager_state_success = self.training_manager.get_state_for_model_checkpoint()
+                state['manager_state'] = manager_state
+
+                if not manager_state_success:
+                    self._log.warn("Getting the manager state for the model checkpoint was not successful, "
+                                   "will continue anyway ...")
+                    success = False
+
+            except Exception as e:
+                _.log_exception(self._log, "Unable to add manager state to model checkpoint, "
+                                           "continuing anyway ...", e)
+                success = False
+        else:
+            # In any case when the state is None, gathering the model checkpoint data is not successful
+            success = False
+
+        return state, success
+
+    def _gather_training_checkpoint_data(self):
+        """
+
+        :return: state, success
+        """
+        return self.training_manager.get_state()
+
+    def _save_model_checkpoint(self, filename, state):
+        with open(filename, 'wb') as f:
+            pickle.dump(state, f)
+
+    def _save_training_checkpoint(self, filename, state):
+        with open(filename, 'wb') as f:
+            pickle.dump(state, f)
 
     def _backup_checkpoint(self, filename):
         if os.path.isfile(filename):
@@ -389,3 +479,12 @@ class CheckpointManagerBase(Callback, metaclass=abc.ABCMeta):
                            f"exact multiple of _create_checkpoint_every, "
                            f"changing archive_last_model_checkpoint_every to "
                            f"[{self._archive_last_model_checkpoint_every}]")
+
+    @staticmethod
+    def pretty_iter_name(iter_name):
+        if iter_name == "epoch":
+            return "Epoch"
+        elif iter_name == "global_iter":
+            return "Global iter"
+        else:
+            return iter_name
