@@ -14,6 +14,7 @@ from mlpug.utils import get_value_at, set_value_at
 
 
 class MetricsLoggingMode(Enum):
+    # TODO : Add BATCH_METRICS_ONLY Option
     BATCH_AND_WINDOW_AVERAGE_METRICS = "BATCH_AND_WINDOW_AVERAGE_METRICS"
     WINDOW_AVERAGE_METRICS = "WINDOW_AVERAGE_METRICS"
     WHOLE_DATASET_METRICS = "WHOLE_DATASET_METRICS"
@@ -38,6 +39,7 @@ class MetricsLoggerBase(Callback):
                  evaluate_settings=None,
                  batch_averaging_window=None,
                  log_condition_func=None,
+                 sliding_window_factory=None,
                  name=None,
                  **kwargs):
         """
@@ -78,9 +80,12 @@ class MetricsLoggerBase(Callback):
         :param batch_averaging_window: Window length in batches, over which the average metric must be calculated
         :type batch_averaging_window: Int
 
+        :param sliding_window_factory: Optional. Function to create SlidingWindow instances
+
         :param name:
         :type name:
         """
+        # print(f"mlpug.MetricsLoggerBase(kwargs={kwargs})")
 
         if name is None:
             name = self.__class__.__name__
@@ -105,6 +110,8 @@ class MetricsLoggerBase(Callback):
 
         self._log_condition_func = log_condition_func or (lambda logs, dataset_batch: True)
 
+        self._sliding_window_factory = SlidingWindow if not callable(sliding_window_factory) else sliding_window_factory
+
         self._name = name
 
         self._metric_windows = {}
@@ -135,7 +142,7 @@ class MetricsLoggerBase(Callback):
             self._log.error('Given state is invalid, unable to set state')
             return False
 
-        self._metric_windows = {metric_path: SlidingWindow(state=window_state)
+        self._metric_windows = {metric_path: self._sliding_window_factory(state=window_state)
                                 for metric_path, window_state in state["metric_windows"].items()}
 
         return True
@@ -147,7 +154,24 @@ class MetricsLoggerBase(Callback):
                           start_batch,
                           start_update_iter):
 
-        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
+        will_log_window_average_metrics = MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode)
+
+        if not self.instance_valid():
+            if will_log_window_average_metrics:
+                self._log.error(f"{self} is not valid, unable to set up averaging window ... ")
+            else:
+                self._log.error(f"{self} is not valid, skipping this hook ... ")
+            return False
+
+        if will_log_window_average_metrics and \
+                (self._batch_averaging_window is None or self._batch_averaging_window <= 0):
+            self._log.error(f"A valid batch processing window is required for "
+                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
+            self._valid = False
+
+            return False
+
+        if will_log_window_average_metrics:
             self._init_metric_windows()
 
         return True
@@ -183,7 +207,9 @@ class MetricsLoggerBase(Callback):
                 set_value_at(base_path, current, dataset_batch_logs)
 
             metric_names = self._metric_evaluator.get_metric_names()
-            metric_paths = get_key_paths(dataset_batch_logs, keys_to_consider=metric_names)
+            metric_paths = get_key_paths(dataset_batch_logs,
+                                         keys_to_consider=metric_names,
+                                         keys_not_to_consider=["auxiliary_results"])
 
             self._update_metrics_windows_for(metric_paths, dataset_batch_logs, base_path=base_path)
 
@@ -279,7 +305,7 @@ class MetricsLoggerBase(Callback):
             if sliding_window is None:
                 self._log.debug(f"Creating sliding window for {full_metric_path}")
 
-                sliding_window = SlidingWindow(length=self._batch_averaging_window, name=full_metric_path)
+                sliding_window = self._sliding_window_factory(length=self._batch_averaging_window, name=full_metric_path)
                 self._metric_windows[metric_path] = sliding_window
 
             sliding_window.slide(metric_value)
@@ -332,13 +358,6 @@ class MetricsLoggerBase(Callback):
                             f"the {self} will not function")
             self._valid = False
 
-        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode) and \
-                (not isinstance(self._batch_averaging_window, int) or
-                 self._batch_averaging_window <= 0):
-            self._log.error(f"A valid batch processing window is required for "
-                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
-            self._valid = False
-
         if self._log_condition_func is not None and not callable(self._log_condition_func):
             self._log.error(f"The log condition function must be callable, "
                             f"metric logging mode ({self._logging_mode}, the {self} will not function")
@@ -359,12 +378,13 @@ class TrainingMetricsLogger(MetricsLoggerBase):
                  logging_mode=MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                  **kwargs):
 
+        # print(f"mlpug.TrainingMetricsLogger(kwargs={kwargs})")
+
         super().__init__(
             metric_evaluator=metric_evaluator,
             dataset_name='training',
             batch_level=batch_level,
             logging_mode=logging_mode,
-            name=name,
             **kwargs)
 
     def on_training_start(self,
@@ -386,7 +406,7 @@ class TrainingMetricsLogger(MetricsLoggerBase):
                 self._valid = False
                 return False
 
-        return self.on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)
+        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)
 
 
 class TestMetricsLogger(MetricsLoggerBase):
@@ -417,6 +437,33 @@ class TestMetricsLogger(MetricsLoggerBase):
 
         self._dataset_iterator = None
 
+    def on_training_start(self,
+                          num_epochs,
+                          num_batches_per_epoch,
+                          start_epoch,
+                          start_batch,
+                          start_update_iter):
+
+        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
+            if self._batch_averaging_window is None:
+                try:
+                    self._batch_averaging_window = len(self._dataset)
+                    if self._batch_averaging_window == math.inf:
+                        self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
+                                        f"unable to calculate window average. "
+                                        f"Please set a finite batch_processing_window during construction of "
+                                        f"this TestMetricsLogger.")
+                        self._valid = False
+                    else:
+                        self._log.debug(f"Set batch averaging window to number of batches in "
+                                        f"{self._dataset_name} data set : {self._batch_averaging_window}")
+                except Exception as e:
+                    _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
+                                               f"{self} will not function", e)
+                    self._valid = False
+
+        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)
+
     def on_epoch_start(self, logs):
         """
 
@@ -457,20 +504,4 @@ class TestMetricsLogger(MetricsLoggerBase):
             self._log.error(f"The given dataset {str(self._dataset)} is not iterable, the {self} will not function")
             self._valid = False
 
-        if self._batch_averaging_window is None:
-            try:
-                self._batch_averaging_window = len(self._dataset)
-                if self._batch_averaging_window == math.inf:
-                    self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
-                                    f"unable to calculate window average. Please set a finite batch_processing_window "
-                                    f"during construction of this TestMetricsLogger.")
-                    self._valid = False
-                else:
-                    self._log.debug(f"Set batch averaging window to number of batches in "
-                                    f"{self._dataset_name} data set : {self._batch_averaging_window}")
-            except Exception as e:
-                _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
-                                           f"{self} will not function", e)
-                self._valid = False
-
         super()._validate()
