@@ -1,6 +1,5 @@
-# TODO : remove PyTorch dependency
 import os
-from torch.utils.tensorboard import SummaryWriter
+from tensorboardX import SummaryWriter
 
 from mlpug.trainers.callbacks.callback import Callback
 from mlpug.utils import get_value_at as get_value_at_func
@@ -17,6 +16,10 @@ def get_value_at(path, obj):
     return get_value_at_func(path, obj, warn_on_failure=False)
 
 
+def tagify(label):
+    return label.replace(' ', '_').lower()
+
+
 class Tensorboard(Callback):
 
     def __init__(self,
@@ -25,6 +28,7 @@ class Tensorboard(Callback):
                  dataset_name=None,
                  label_name=None,
                  batch_level=True,
+                 write_on_epoch_complete=False,
                  metrics_are_averages=False,
                  metric_names=None,
                  batch_log_period=1,
@@ -32,7 +36,8 @@ class Tensorboard(Callback):
                  log_dir='../training-logs/',
                  tensorboard_options=None,
                  ignore_missing_metrics=False,
-                 name="Tensorboard"):
+                 name="Tensorboard",
+                 **kwargs):
         """
 
         Shows batch-level or epoch-level metrics, as defined by the metric key paths in `metric_paths`.
@@ -159,7 +164,6 @@ class Tensorboard(Callback):
         Further, the `hamburger` and `hotdog` precision plots will be combined in one precision plot.
 
 
-
         ## Using `metrics_are_averages` and `metric_names`
 
         If `metrics_are_averages=True` the metrics are written to Tensorboard using another writer, such that
@@ -213,6 +217,10 @@ class Tensorboard(Callback):
 
         :param batch_level:     {boolean} If True, the given metrics will be logged per batch, else per epoch
 
+        :param write_on_epoch_complete: {boolean} If True, and batch_level = True, logs will also be written
+                                        on_epoch_complete. This is useful when on_epoch_complete metrics are calculated
+                                        that you want to incorporate in a batch_level graph.
+
         :param metrics_are_averages: {boolean}
                                      If True, the given metrics are written by a separate writer dedicated
                                      to metrics window averages
@@ -248,7 +256,7 @@ class Tensorboard(Callback):
         :param name:
         """
 
-        super().__init__(name=name)
+        super().__init__(name=name, **kwargs)
 
         self._metric_paths = metric_paths
 
@@ -257,6 +265,7 @@ class Tensorboard(Callback):
         self._label_name = label_name
 
         self._batch_level = batch_level
+        self._write_on_epoch_complete = write_on_epoch_complete
         self._metrics_are_averages = metrics_are_averages
 
         self._metric_names = metric_names
@@ -320,14 +329,18 @@ class Tensorboard(Callback):
             return False
 
         if self._batch_level:
-            return True
+            if self._write_on_epoch_complete:
+                # There are dataset level metrics calculated that are followed on batch level.
+                return self.on_batch_training_completed(None, logs)
+            else:
+                return True
 
         current = self._get_logs_base(logs)
         epoch = current['epoch']
 
         return self._write(current, epoch)
 
-    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
+    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
         self._writer.close()
         return True
 
@@ -387,8 +400,13 @@ class Tensorboard(Callback):
     def _write(self, current_logs, training_iter):
         metrics, success = self._get_metrics_from(current_logs)
 
-        for tag, metric in metrics.items():
-            self._writer.add_scalar(tag, metric, global_step=training_iter)
+        for label, metric in metrics.items():
+            if type(metric) is tuple:
+                # use the first value as metric value, the other values are auxiliary results meant for other purposes
+                metric = metric[0]
+
+            tag = tagify(label)
+            self._writer.add_scalar(tag, metric, global_step=training_iter, display_name=label)
 
         if training_iter % self._flush_period == 0:
             self._writer.flush()
@@ -410,8 +428,8 @@ class Tensorboard(Callback):
             if metric is None:
                 return None
 
-            tag = self._get_tag(metric_path)
-            metrics[tag] = metric
+            label = self._get_label(metric_path)
+            metrics[label] = metric
 
             return metric
 
@@ -431,14 +449,14 @@ class Tensorboard(Callback):
 
         return metrics, success
 
-    def _get_tag(self, metric_name):
+    def _get_label(self, metric_name):
         iter_level = 'batch' if self._batch_level else 'epoch'
 
-        tag = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
-        tag = tag.replace('.', '_')
-        tag = f"{tag} - per {iter_level}"
+        label = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
+        label = label.replace('.', ' ').replace('_', ' ').title()
+        label = f"{label} - per {iter_level}"
 
-        return tag
+        return label
 
 
 class AutoTensorboard(Callback):
@@ -454,7 +472,9 @@ class AutoTensorboard(Callback):
                  flush_period=50,
                  log_dir='../training-logs/',
                  tensorboard_options=None,
-                 name="AutoTensorboard"):
+                 ignore_missing_metrics=False,
+                 name="AutoTensorboard",
+                 **kwargs):
         """
 
         Shows batch-level metrics, and/or averaged batch-level (metrics), and/or epoch level metrics.
@@ -540,9 +560,11 @@ class AutoTensorboard(Callback):
                                     Optional dictionary with options that need to be used when instantiating the
                                     Tensorboard writers.
 
+        :param :param ignore_missing_metrics {Boolean}
+
         :param name:
         """
-        super().__init__(name=name)
+        super().__init__(name=name, **kwargs)
 
         self._experiment_name = experiment_name
         self._dataset_name = dataset_name
@@ -561,6 +583,8 @@ class AutoTensorboard(Callback):
 
         self._tensorboard_options = tensorboard_options or {}
 
+        self._ignore_missing_metrics = ignore_missing_metrics
+
         self._writers = dict()
 
     def _setup(self):
@@ -616,14 +640,14 @@ class AutoTensorboard(Callback):
             return False
 
         if not self._show_epoch_level:
-            return
+            return True
 
         current = self._get_logs_base(logs)
         epoch = current['epoch']
 
         return self._write(current, epoch, writer_type=METRIC_WRITER, batch_level=False)
 
-    def on_training_ended(self, stopped_early, stopped_on_error, callback_calls_success):
+    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
         for writer in self._writers.values():
             writer.close()
         return True
@@ -666,8 +690,21 @@ class AutoTensorboard(Callback):
     def _write(self, current_logs, training_iter, writer_type, batch_level):
         metrics, success = self._get_metrics_from(current_logs, writer_type, batch_level)
 
-        for tag, metric in metrics.items():
-            self._writers[writer_type].add_scalar(tag, metric, global_step=training_iter)
+        if not success:
+            if self._ignore_missing_metrics:
+                return True
+            else:
+                self._log.error("No metrics found, nothing to write to Tensorboard")
+                return success
+
+        for label, metric in metrics.items():
+            if type(metric) is tuple:
+                # use the first value as metric value, the other values are auxiliary results meant for other purposes
+                metric = metric[0]
+
+            tag = tagify(label)
+
+            self._writers[writer_type].add_scalar(tag, metric, global_step=training_iter, display_name=label)
 
         if training_iter % self._flush_period == 0:
             self._writers[writer_type].flush()
@@ -690,8 +727,9 @@ class AutoTensorboard(Callback):
                 alt_base_path = f"{self._dataset_name}.window_average"
 
             metrics, success = self._get_all_metrics_from(base_path, current_logs, batch_level)
-            if alt_base_path is not None:
-                self._log.warn(f'Trying alternate metrics path {alt_base_path} ...')
+            if not success and alt_base_path is not None:
+                if not self._ignore_missing_metrics:
+                    self._log.warn(f'Trying alternate metrics path {alt_base_path} ...')
                 alt_metrics, alt_success = self._get_all_metrics_from(alt_base_path, current_logs, batch_level)
 
                 if alt_success:
@@ -733,9 +771,9 @@ class AutoTensorboard(Callback):
                     _add_metrics(metric, metric_name)
                     continue
 
-                tag = self._get_tag(metric_name, batch_level)
+                label = self._get_label(metric_name, batch_level)
 
-                metrics[tag] = metric
+                metrics[label] = metric
 
         if _.is_dict(dataset_metrics):
             _add_metrics(dataset_metrics)
@@ -743,17 +781,19 @@ class AutoTensorboard(Callback):
             if len(metrics) > 0:
                 return metrics, True
             else:
-                self._log.warn(f'No metrics found at {base_path}')
+                if not self._ignore_missing_metrics:
+                    self._log.warn(f'No metrics found at {base_path}')
                 return metrics, False
         else:
-            self._log.error(f'No valid metrics found for {base_path}')
+            if not self._ignore_missing_metrics:
+                self._log.error(f'No valid metrics found for {base_path}')
             return None, False
 
-    def _get_tag(self, metric_name, batch_level):
+    def _get_label(self, metric_name, batch_level):
         iter_level = 'batch' if batch_level else 'epoch'
 
-        tag = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
-        tag = tag.replace('.', '_')
-        tag = f"{tag} - per {iter_level}"
+        label = self._metric_names[metric_name] if (metric_name in self._metric_names) else metric_name
+        label = label.replace('.', ' ').replace('_', ' ').title()
+        label = f"{label} - per {iter_level}"
 
-        return tag
+        return label
