@@ -5,7 +5,7 @@ import time
 
 from statistics import mean
 
-from basics.base import Base
+from mlpug.base import Base
 import basics.base_utils as _
 
 from mlpug.utils import convert_to_dict, get_value_at, set_value_at, has_key, SlidingWindow
@@ -13,6 +13,24 @@ from mlpug.utils import convert_to_dict, get_value_at, set_value_at, has_key, Sl
 from mlpug.mlpug_exceptions import TrainerInvalidException
 
 
+def normalize_evaluation(results):
+    if type(results) is dict:
+        return results
+    elif type(results) is tuple:
+        return {
+            "loss": results[0],
+            "auxiliary_results": results[1:]
+        }
+    else:
+        return {
+            "loss": results
+        }
+
+
+class BatchChunkingResults(list):
+    pass
+
+
 class TrainingManager(Base, metaclass=abc.ABCMeta):
 
     def __init__(self,
@@ -21,7 +39,9 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                  num_epochs=50,
                  num_batches_per_epoch=None,
                  callbacks=None,
-                 experiment_data=None):
+                 experiment_data=None,
+                 sliding_window_factory=None,
+                 **kwargs):
         """
         Implements a simple training loop in the `_train` method. Although simple in nature, the use of callbacks
         (see the `Callback` class), provide the means to monitor, control and customize the training.
@@ -40,9 +60,10 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
         :param callbacks:
         :param experiment_data: Data object that will be added to the training manager state.
                                 This could contain additional training/model experiment data
+        :param sliding_window_factory: Optional. Function to create SlidingWindow instances
 
         """
-        super(TrainingManager, self).__init__()
+        super(TrainingManager, self).__init__(**kwargs)
 
         self.trainer = trainer
 
@@ -61,8 +82,12 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         self.experiment_data = experiment_data
 
+        self.sliding_window_factory = SlidingWindow if not callable(sliding_window_factory) else sliding_window_factory
+
         self.init_logs = None
-        self.logs = {}
+        self.logs = {
+            "cb_calls_success": True
+        }
 
         self._previous_batch_duration = None
 
@@ -70,22 +95,27 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         self._stop_training = False
 
-        self._callback_calls_successful = True
-
         if not self._validate():
             return
 
         self._assess_num_batches_per_epoch()
 
-        if type(self.num_batches_per_epoch) == int and self.num_batches_per_epoch > 0:
-            self._metric_windows['training.batch.loss'] = \
-                SlidingWindow(length=self.num_batches_per_epoch, name='training.batch.loss')
+        try:
+            # TODO : better deal with num_batches_per_epoch special cases (e.g. Numpy number)
+            # TODO : deal with the case when we don't have SlidingWindows
+            if float(self.num_batches_per_epoch) != math.inf and self.num_batches_per_epoch > 0:
+                self._metric_windows['training_params.batch.duration'] = \
+                    self.sliding_window_factory(length=self.num_batches_per_epoch,
+                                                name='training_params.batch.duration')
 
-            self._metric_windows['training_params.batch.duration'] = \
-                SlidingWindow(length=self.num_batches_per_epoch, name='training_params.batch.duration')
+        except Exception as e:
+            self._valid = False
+            _.log_exception(self._log,
+                            "Unable to setup sliding window for training loss and batch training duration", e)
 
         if not self._call_callbacks("set_training_manager", self):
-            self._callback_calls_successful = False
+            self._valid = False
+            self.logs["cb_calls_success"] = False
             self._log.error("One or more issues occurred while providing this training manager instance "
                             "to the provided callbacks")
 
@@ -100,12 +130,21 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
         try:
             self._train()
         except KeyboardInterrupt:
-            sys.stdout.write("\n\n")
+            self._update_cb_success(self._call_callbacks('on_training_ended',
+                                                         stopped_early=False,
+                                                         stopped_on_error=False,
+                                                         interrupted=True,
+                                                         callback_calls_success=self.logs["cb_calls_success"]))
+
+            sys.stdout.write("\n")
             self._log.info('Training process interrupted by you ... 🤷🏻‍♂️\n')
-            # TODO : deal with the distributed training case, wait to return until
-            #  all workers have stopped training
+            sys.stdout.flush()
+
             self.stop_training()
 
+        self._training_ended()
+
+
     def stop_training(self):
         self._stop_training = True
 
@@ -276,11 +315,11 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
             if self.num_batches_per_epoch and (window_length != self.num_batches_per_epoch):
                 # override when different number of batches per epoch is given (or calculated)
                 # during construction
-                window = SlidingWindow(length=self.num_batches_per_epoch,
-                                       init_window_values=window_state['window'],
-                                       name=metric_path)
+                window = self.sliding_window_factory(length=self.num_batches_per_epoch,
+                                                     init_window_values=window_state['window'],
+                                                     name=metric_path)
             else:
-                window = SlidingWindow(state=window_state)
+                window = self.sliding_window_factory(state=window_state)
 
             self._metric_windows[metric_path] = window
         except Exception as e:
@@ -319,14 +358,27 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
         epoch_stopped_early = False
         training_stopped_on_error = False
 
-        cb_calls_success = self._callback_calls_successful
+        if self.init_logs is not None:
+            self._log.debug("Setting provided initial log object ...")
 
-        cb_calls_success &= self._call_callbacks('on_training_start',
-                                                 self.num_epochs,
-                                                 self.num_batches_per_epoch,
-                                                 self.epoch,
-                                                 self.batch_step,
-                                                 self.global_iter)
+            cb_calls_success = self.logs["cb_calls_success"]
+            self.logs = self.init_logs
+            self.logs["cb_calls_success"] &= cb_calls_success
+
+            self.init_logs = None
+
+        # override log state for certain parameters given at construction
+        self.logs["final_epoch"] = self.num_epochs - 1
+        self.logs["final_batch_step"] = self.num_batches_per_epoch - 1
+
+        update = self._update_cb_success
+
+        update(self._call_callbacks('on_training_start',
+                                    self.num_epochs,
+                                    self.num_batches_per_epoch,
+                                    self.epoch,
+                                    self.batch_step,
+                                    self.global_iter))
 
         for epoch in range(self.epoch, self.num_epochs):
             self.epoch = epoch
@@ -336,38 +388,19 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                 self._log.warn("Training stopped early")
                 break
 
-            if self.init_logs is not None:
-                self._log.debug("Setting provided initial logs object ...")
-                self.logs = self.init_logs
-
-                # override log state for certain parameters given at construction
-                self.logs["final_epoch"] = self.num_epochs-1
-                self.logs["final_batch_step"] = self.num_batches_per_epoch - 1
-                self.init_logs = None
-            else:
-                # Keep anything in the logs from the previous epoch that is non-standard
-                self.logs = self.logs or {}
-                self.logs = {**self.logs, **{
-                    "final_epoch": self.num_epochs - 1,
-                    "final_batch_step": self.num_batches_per_epoch - 1,
-
-                    "epoch": self.epoch,
-
-                    "current": None,
-
-                    "cb_calls_success": cb_calls_success
-                }}
-
-            # It's a bit shorter
-            logs = self.logs
-
-            logs["cb_calls_success"] &= self._call_callbacks('on_epoch_start', logs)
-
-            epoch_stopped_early = False
+            # Keep anything in the logs from the previous epoch that is non-standard
+            # Reset all other values
+            self.logs = {**self.logs, **{
+                "epoch": self.epoch,
+                "current": None,
+            }}
 
             training_dataset = self._prepare_training_dataset()
+
+            logs = self.logs  # just a bit shorter
             current = None
-            ctl = None
+            epoch_started = True
+            epoch_stopped_early = False
             for training_batch in iter(training_dataset):
                 batch_training_start_time = time.time()
 
@@ -377,20 +410,24 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                     self._log.warn("Training epoch stopped early")
                     break
 
-                logs["current"] = self._init_current_logs()
+                current = self._init_current_logs()
+                logs["current"] = current
 
-                logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_start', training_batch, logs)
-
-                current = logs["current"]
-                ctl = current["training"]
+                # on_epoch_start needs to be called in the loop, because some callbacks might need the current object
+                if epoch_started:
+                    update(self._call_callbacks('on_epoch_start', self.logs))
+                    epoch_started = False
 
+                update(self._call_callbacks('on_batch_training_start', training_batch, logs))
                 try:
-                    training_settings = logs["training_settings"] if has_key(logs, "training_settings") else None
-                    ctl["batch"]["loss"], ctl["batch"]["auxiliary_results"] = self.trainer.train_on(training_batch,
-                                                                                                    training_settings)
+                    # Note: Tensorflow needs a defined training_settings
+                    if not has_key(logs, "training_settings"):
+                        logs["training_settings"] = {}
 
-                    self._update_window("training.batch.loss")
-                    ctl["window_average"]["loss"] = self._calc_window_average("training.batch.loss")
+                    loss, auxiliary_results = self.trainer.train_on(training_batch, logs["training_settings"])
+
+                    set_value_at("training.batch.loss", current, loss)
+                    set_value_at("training.batch.auxiliary_results", current, auxiliary_results)
                 except Exception as e:
                     if isinstance(e, TrainerInvalidException):
                         err_msg = f"Trainer {self.trainer} is misconfigured, unable to train on batch, " \
@@ -401,23 +438,21 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
                     _.log_exception(self._log, err_msg, e)
 
-                    logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_failed', e, logs)
+                    update(self._call_callbacks('on_batch_training_failed', e, logs))
                     training_stopped_on_error = True
 
                 if training_stopped_on_error:
                     break
 
-                logs["cb_calls_success"] &= self._call_callbacks('on_batch_training_completed', training_batch, logs)
+                update(self._call_callbacks('on_batch_training_completed', training_batch, logs))
 
                 self.batch_step += 1
                 self.global_iter += 1
 
                 batch_training_end_time = time.time()
-
                 batch_duration = batch_training_end_time - batch_training_start_time
 
                 set_value_at("training_params.batch.duration", current, batch_duration)
-
                 self._update_window("training_params.batch.duration")
 
                 self._previous_batch_duration = batch_duration
@@ -429,27 +464,27 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
                     break
 
             if not (epoch_stopped_early or training_stopped_on_error):
-                ctpl = current["training_params"]
+                mean_batch_duration = self._calc_window_average("training_params.batch.duration")
+                epoch_duration = self._calc_window_sum("training_params.batch.duration")
 
-                ctpl["window_average"]["duration"] = self._calc_window_average("training_params.batch.duration")
-                ctpl["epoch"]["duration"] = self._calc_window_sum("training_params.batch.duration")
+                set_value_at("training_params.window_average.duration", current, mean_batch_duration)
+                set_value_at("training_params.epoch.duration", current, epoch_duration)
 
-                logs["cb_calls_success"] &= self._call_callbacks('on_epoch_completed', logs)
-
-            cb_calls_success &= logs["cb_calls_success"]
+                update(self._call_callbacks('on_epoch_completed', logs))
 
             if training_stopped_on_error:
                 break
 
             self.batch_step = 0
 
-        cb_calls_success &= self._call_callbacks('on_training_ended',
-                                                 stopped_early=epoch_stopped_early or training_stopped_early,
-                                                 stopped_on_error=training_stopped_on_error,
-                                                 callback_calls_success=cb_calls_success)
+        update(self._call_callbacks('on_training_ended',
+                                    stopped_early=epoch_stopped_early or training_stopped_early,
+                                    stopped_on_error=training_stopped_on_error,
+                                    interrupted=False,
+                                    callback_calls_success=self.logs["cb_calls_success"]))
 
         if not training_stopped_on_error:
-            if cb_calls_success:
+            if self.logs["cb_calls_success"]:
                 self._log.info("Training completed. All good! ❤️")
             else:
                 self._log.warn("Training completed, but one or more callback errors occurred. "
@@ -465,9 +500,7 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
             "training": {
                 "batch": {},
-                "window_average": {
-                    "loss": self._calc_window_average("training.batch.loss")
-                },
+                "window_average": {},
                 "dataset": {},
             },
 
@@ -504,6 +537,9 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
         return success
 
+    def _update_cb_success(self, cb_calls_success):
+        self.logs["cb_calls_success"] &= cb_calls_success
+
     def _update_window(self, metric_path):
         try:
             window = self._metric_windows[metric_path]
@@ -563,6 +599,9 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
         except Exception as e:
             _.log_exception(self._log, err_msg, e)
 
+    def _training_ended(self):
+        pass
+
     def _validate(self):
         # TODO
         self._valid = True
@@ -572,14 +611,14 @@ class TrainingManager(Base, metaclass=abc.ABCMeta):
 
 class TrainerBase(Base, metaclass=abc.ABCMeta):
 
-    def __init__(self, model_components, optimizers):
+    def __init__(self, model_components, optimizers, name="TrainerBase", **kwargs):
         """
 
         :param model_components: dict with model components
         :param optimizers: dict with optimizers
 
         """
-        super(TrainerBase, self).__init__()
+        super(TrainerBase, self).__init__(pybase_logger_name=name, **kwargs)
 
         self.model_components = model_components
         self.optimizers = optimizers
@@ -751,7 +790,7 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
 
         :return: True on success, else False
         """
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def train_on(self, batch_data, training_settings=None):
@@ -769,7 +808,7 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
         loss : number (e.g. float)
         auxiliary_results : can be anything, e.g dict or list with values or data items
         """
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     def evaluate_loss(self, batch_data, inference_mode, evaluate_settings=None):
         """
@@ -789,43 +828,53 @@ class TrainerBase(Base, metaclass=abc.ABCMeta):
 
         self._activate_inference_mode(inference_mode)
 
-        return self._evaluate_loss(batch_data, evaluate_settings)
+        if evaluate_settings is None:
+            evaluate_settings = {}
+
+        results = self._evaluate_loss(batch_data, evaluate_settings, inference_mode)
+        return normalize_evaluation(results)
 
     @abc.abstractmethod
-    def _evaluate_loss(self, batch_data, evaluate_settings=None):
+    def _evaluate_loss(self, batch_data, evaluate_settings=None, inference_mode=None):
         """
+        Evaluates the given training model on the  given batch_data, using the optional training_settings
+        Depending on the Deep learning backend you might need to use inference mode here
 
         :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
         :param evaluate_settings: optional evaluate_settings object (usually dict)
+        :param inference_mode: bool, important when inference mode not set in `_activate_inference_mode`
+                               Pytorch:     inference_mode not required here
+                               Tensorflow:  inference_mode required here
 
-        :return: dict:
+        :return: dict or tuple
             {
                 "loss": <Tensor>,
                 "auxiliary_results": <can be anything, e.g dict or list with values or data items>
             }
-        """
 
-        pass
+            (loss, ... auxiliary results ...)
+        """
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def _activate_inference_mode(self, inference_mode):
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def _get_model_state(self, model, model_name=None):
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def _get_optimizer_state(self, optimizer, optimizer_name=None):
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def _set_model_state(self, model, state, model_name=None):
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     @abc.abstractmethod
     def _set_optimizer_state(self, optimizer, state, optimizer_name):
-        pass
+        raise NotImplemented("Please implement this method in your child class")
 
     def _check_state(self, state):
         state_attributes = ['model_components', 'optimizers']
@@ -851,12 +900,15 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
                  optimizers,
                  model_components=None,
                  batch_chunk_size=None,
-                 use_mixed_precision=False):
+                 use_mixed_precision=False,
+                 name="DefaultTrainerBase",
+                 **kwargs):
         """
         Simple trainer based on a training_model, that evaluates the loss on batch data
 
         :param optimizers: dict or list with optimizer(s), or a single optimizer instance
         :param model_components: dict or list with model components(s), or a single model instance
+
         :param batch_chunk_size: optional batch chunk size (int)
                                  If given, batches are processed in chunks of size `batch_chunk_size` samples to
                                  calculate the gradients. The last chunk can be smaller than `batch_chunk_size` if
@@ -877,7 +929,7 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
         model_components = convert_to_dict("model", model_components)
         optimizers = convert_to_dict("optimizer", optimizers)
 
-        super().__init__(model_components, optimizers)
+        super().__init__(model_components, optimizers, name=name, **kwargs)
 
         self.batch_chunk_size = batch_chunk_size
         if self.batch_chunk_size is not None:
@@ -899,7 +951,10 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
                         "auxiliary_results": <can be anything, e.g dict or list with values or data items>
                       }
 
-        :param model:
+                      or
+
+                      model(self, batch_data, training_settings) -> [<loss_tensor>, ... auxiliary_results ...]
+
         :return:
         """
 
@@ -908,22 +963,27 @@ class DefaultTrainerBase(TrainerBase, metaclass=abc.ABCMeta):
         instance_valid = self.instance_valid()
         self._valid = self._validate_model() & instance_valid
 
-    def _evaluate_loss(self, batch_data, evaluate_settings=None):
+    def _evaluate_loss(self, batch_data, evaluate_settings=None, inference_mode=None):
         """
-
         Evaluates the given training model on the  given batch_data, using the optional training_settings
+        Depending on the Deep learning backend you might need to use inference mode here
 
         :param batch_data: batch_data object to evaluate loss on (e.g. dict, list, tuple)
         :param evaluate_settings: optional evaluate_settings object (usually dict)
+        :param inference_mode: optional bool, important when inference mode not set in `_activate_inference_mode`
+                               Pytorch:     inference_mode not required here
+                               Tensorflow:  inference_mode required here
 
-
-        :return: dict:
+        :return: dict or tuple
             {
                 "loss": <Tensor>,
                 "auxiliary_results": <can be anything, e.g dict or list with values or data items>
             }
+
+            (loss, ... auxiliary results ...)
         """
-        return self.training_model(batch_data, evaluate_settings)
+
+        return self.training_model(batch_data, evaluate_settings, inference_mode)
 
     def _validate_model(self):
         # TODO : this is framework dependent
